"""
THIS FILE IS GENERATED. DO NOT EDIT IT MANUALLY.
"""
from __future__ import annotations
from functools import cache
from sqlite3 import Connection
from pathlib import Path

class ResourcePack:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.ResourcePack_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT path
            FROM ResourcePack
            WHERE ResourcePack_pk = ?
            ''',
            (self.ResourcePack_pk,)
        ).fetchone()

    @property
    def path(self) -> Path:
        return self.query_result()[0]


class ClientEntityFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.ClientEntityFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ResourcePack_fk, path
            FROM ClientEntityFile
            WHERE ClientEntityFile_pk = ?
            ''',
            (self.ClientEntityFile_pk,)
        ).fetchone()

    @property
    def ResourcePack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class ClientEntity:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.ClientEntity_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ClientEntityFile_fk, identifier
            FROM ClientEntity
            WHERE ClientEntity_pk = ?
            ''',
            (self.ClientEntity_pk,)
        ).fetchone()

    @property
    def ClientEntityFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]


class ClientEntityRenderControllerField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.ClientEntityRenderControllerField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ClientEntity_fk, condition, identifier, jsonPath
            FROM ClientEntityRenderControllerField
            WHERE ClientEntityRenderControllerField_pk = ?
            ''',
            (self.ClientEntityRenderControllerField_pk,)
        ).fetchone()

    @property
    def ClientEntity_fk(self) -> int:
        return self.query_result()[0]

    @property
    def condition(self) -> str | None:
        return self.query_result()[1]

    @property
    def identifier(self) -> str:
        return self.query_result()[2]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[3]


class ClientEntityGeometryField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.ClientEntityGeometryField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ClientEntity_fk, identifier, jsonPath, shortName
            FROM ClientEntityGeometryField
            WHERE ClientEntityGeometryField_pk = ?
            ''',
            (self.ClientEntityGeometryField_pk,)
        ).fetchone()

    @property
    def ClientEntity_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def shortName(self) -> str:
        return self.query_result()[3]


class ClientEntityTextureField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.ClientEntityTextureField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ClientEntity_fk, identifier, jsonPath, shortName
            FROM ClientEntityTextureField
            WHERE ClientEntityTextureField_pk = ?
            ''',
            (self.ClientEntityTextureField_pk,)
        ).fetchone()

    @property
    def ClientEntity_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def shortName(self) -> str:
        return self.query_result()[3]


class ClientEntityMaterialField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.ClientEntityMaterialField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ClientEntity_fk, identifier, jsonPath, shortName
            FROM ClientEntityMaterialField
            WHERE ClientEntityMaterialField_pk = ?
            ''',
            (self.ClientEntityMaterialField_pk,)
        ).fetchone()

    @property
    def ClientEntity_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def shortName(self) -> str:
        return self.query_result()[3]


class ClientEntityAnimationField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.ClientEntityAnimationField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ClientEntity_fk, identifier, jsonPath, shortName
            FROM ClientEntityAnimationField
            WHERE ClientEntityAnimationField_pk = ?
            ''',
            (self.ClientEntityAnimationField_pk,)
        ).fetchone()

    @property
    def ClientEntity_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def shortName(self) -> str:
        return self.query_result()[3]


class ClientEntityAnimationControllerField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.ClientEntityAnimationControllerField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ClientEntity_fk, identifier, jsonPath, shortName
            FROM ClientEntityAnimationControllerField
            WHERE ClientEntityAnimationControllerField_pk = ?
            ''',
            (self.ClientEntityAnimationControllerField_pk,)
        ).fetchone()

    @property
    def ClientEntity_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def shortName(self) -> str:
        return self.query_result()[3]


class RenderControllerFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RenderControllerFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ResourcePack_fk, path
            FROM RenderControllerFile
            WHERE RenderControllerFile_pk = ?
            ''',
            (self.RenderControllerFile_pk,)
        ).fetchone()

    @property
    def ResourcePack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class RenderController:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RenderController_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT RenderControllerFile_fk, identifier, jsonPath
            FROM RenderController
            WHERE RenderController_pk = ?
            ''',
            (self.RenderController_pk,)
        ).fetchone()

    @property
    def RenderControllerFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]


class RenderControllerTexturesField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RenderControllerTexturesField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT RenderController_fk, inOwnerArrayJsonPath, jsonPath, ownerArray, shortName
            FROM RenderControllerTexturesField
            WHERE RenderControllerTexturesField_pk = ?
            ''',
            (self.RenderControllerTexturesField_pk,)
        ).fetchone()

    @property
    def RenderController_fk(self) -> int:
        return self.query_result()[0]

    @property
    def inOwnerArrayJsonPath(self) -> str | None:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def ownerArray(self) -> str | None:
        return self.query_result()[3]

    @property
    def shortName(self) -> str:
        return self.query_result()[4]


class RenderControllerMaterialsField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RenderControllerMaterialsField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT RenderController_fk, boneNamePattern, inOwnerArrayJsonPath, jsonPath, ownerArray, shortName
            FROM RenderControllerMaterialsField
            WHERE RenderControllerMaterialsField_pk = ?
            ''',
            (self.RenderControllerMaterialsField_pk,)
        ).fetchone()

    @property
    def RenderController_fk(self) -> int:
        return self.query_result()[0]

    @property
    def boneNamePattern(self) -> str | None:
        return self.query_result()[1]

    @property
    def inOwnerArrayJsonPath(self) -> str | None:
        return self.query_result()[2]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[3]

    @property
    def ownerArray(self) -> str | None:
        return self.query_result()[4]

    @property
    def shortName(self) -> str:
        return self.query_result()[5]


class RenderControllerGeometryField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RenderControllerGeometryField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT RenderController_fk, inOwnerArrayJsonPath, jsonPath, ownerArray, shortName
            FROM RenderControllerGeometryField
            WHERE RenderControllerGeometryField_pk = ?
            ''',
            (self.RenderControllerGeometryField_pk,)
        ).fetchone()

    @property
    def RenderController_fk(self) -> int:
        return self.query_result()[0]

    @property
    def inOwnerArrayJsonPath(self) -> str | None:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def ownerArray(self) -> str | None:
        return self.query_result()[3]

    @property
    def shortName(self) -> str:
        return self.query_result()[4]


class GeometryFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.GeometryFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ResourcePack_fk, path
            FROM GeometryFile
            WHERE GeometryFile_pk = ?
            ''',
            (self.GeometryFile_pk,)
        ).fetchone()

    @property
    def ResourcePack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class Geometry:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.Geometry_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT GeometryFile_fk, identifier, jsonPath, parent
            FROM Geometry
            WHERE Geometry_pk = ?
            ''',
            (self.Geometry_pk,)
        ).fetchone()

    @property
    def GeometryFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def parent(self) -> str | None:
        return self.query_result()[3]


class TextureFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.TextureFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ResourcePack_fk, identifier, path
            FROM TextureFile
            WHERE TextureFile_pk = ?
            ''',
            (self.TextureFile_pk,)
        ).fetchone()

    @property
    def ResourcePack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def path(self) -> Path:
        return self.query_result()[2]


class ParticleFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.ParticleFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ResourcePack_fk, path
            FROM ParticleFile
            WHERE ParticleFile_pk = ?
            ''',
            (self.ParticleFile_pk,)
        ).fetchone()

    @property
    def ResourcePack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class Particle:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.Particle_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ParticleFile_fk, identifier, material, texture
            FROM Particle
            WHERE Particle_pk = ?
            ''',
            (self.Particle_pk,)
        ).fetchone()

    @property
    def ParticleFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def material(self) -> str | None:
        return self.query_result()[2]

    @property
    def texture(self) -> str | None:
        return self.query_result()[3]


class RpAnimationFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RpAnimationFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ResourcePack_fk, path
            FROM RpAnimationFile
            WHERE RpAnimationFile_pk = ?
            ''',
            (self.RpAnimationFile_pk,)
        ).fetchone()

    @property
    def ResourcePack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class RpAnimation:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RpAnimation_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT RpAnimationFile_fk, identifier, jsonPath
            FROM RpAnimation
            WHERE RpAnimation_pk = ?
            ''',
            (self.RpAnimation_pk,)
        ).fetchone()

    @property
    def RpAnimationFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]


class RpAnimationParticleEffect:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RpAnimationParticleEffect_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT RpAnimation_fk, jsonPath, shortName
            FROM RpAnimationParticleEffect
            WHERE RpAnimationParticleEffect_pk = ?
            ''',
            (self.RpAnimationParticleEffect_pk,)
        ).fetchone()

    @property
    def RpAnimation_fk(self) -> int:
        return self.query_result()[0]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[1]

    @property
    def shortName(self) -> str:
        return self.query_result()[2]


class RpAnimationSoundEffect:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RpAnimationSoundEffect_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT RpAnimation_fk, jsonPath, shortName
            FROM RpAnimationSoundEffect
            WHERE RpAnimationSoundEffect_pk = ?
            ''',
            (self.RpAnimationSoundEffect_pk,)
        ).fetchone()

    @property
    def RpAnimation_fk(self) -> int:
        return self.query_result()[0]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[1]

    @property
    def shortName(self) -> str:
        return self.query_result()[2]


class RpAnimationControllerFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RpAnimationControllerFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ResourcePack_fk, path
            FROM RpAnimationControllerFile
            WHERE RpAnimationControllerFile_pk = ?
            ''',
            (self.RpAnimationControllerFile_pk,)
        ).fetchone()

    @property
    def ResourcePack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class RpAnimationController:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RpAnimationController_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT RpAnimationControllerFile_fk, identifier, jsonPath
            FROM RpAnimationController
            WHERE RpAnimationController_pk = ?
            ''',
            (self.RpAnimationController_pk,)
        ).fetchone()

    @property
    def RpAnimationControllerFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]


class RpAnimationControllerParticleEffect:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RpAnimationControllerParticleEffect_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT RpAnimationController_fk, jsonPath, shortName
            FROM RpAnimationControllerParticleEffect
            WHERE RpAnimationControllerParticleEffect_pk = ?
            ''',
            (self.RpAnimationControllerParticleEffect_pk,)
        ).fetchone()

    @property
    def RpAnimationController_fk(self) -> int:
        return self.query_result()[0]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[1]

    @property
    def shortName(self) -> str:
        return self.query_result()[2]


class RpAnimationControllerSoundEffect:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RpAnimationControllerSoundEffect_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT RpAnimationController_fk, jsonPath, shortName
            FROM RpAnimationControllerSoundEffect
            WHERE RpAnimationControllerSoundEffect_pk = ?
            ''',
            (self.RpAnimationControllerSoundEffect_pk,)
        ).fetchone()

    @property
    def RpAnimationController_fk(self) -> int:
        return self.query_result()[0]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[1]

    @property
    def shortName(self) -> str:
        return self.query_result()[2]


class AttachableFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.AttachableFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ResourcePack_fk, path
            FROM AttachableFile
            WHERE AttachableFile_pk = ?
            ''',
            (self.AttachableFile_pk,)
        ).fetchone()

    @property
    def ResourcePack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class Attachable:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.Attachable_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT AttachableFile_fk, identifier
            FROM Attachable
            WHERE Attachable_pk = ?
            ''',
            (self.Attachable_pk,)
        ).fetchone()

    @property
    def AttachableFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]


class AttachableItemField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.AttachableItemField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT Attachable_fk, condition, identifier, jsonPath
            FROM AttachableItemField
            WHERE AttachableItemField_pk = ?
            ''',
            (self.AttachableItemField_pk,)
        ).fetchone()

    @property
    def Attachable_fk(self) -> int:
        return self.query_result()[0]

    @property
    def condition(self) -> str | None:
        return self.query_result()[1]

    @property
    def identifier(self) -> str:
        return self.query_result()[2]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[3]


class AttachableMaterialField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.AttachableMaterialField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT Attachable_fk, identifier, jsonPath, shortName
            FROM AttachableMaterialField
            WHERE AttachableMaterialField_pk = ?
            ''',
            (self.AttachableMaterialField_pk,)
        ).fetchone()

    @property
    def Attachable_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def shortName(self) -> str:
        return self.query_result()[3]


class AttachableTextureField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.AttachableTextureField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT Attachable_fk, identifier, jsonPath, shortName
            FROM AttachableTextureField
            WHERE AttachableTextureField_pk = ?
            ''',
            (self.AttachableTextureField_pk,)
        ).fetchone()

    @property
    def Attachable_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def shortName(self) -> str:
        return self.query_result()[3]


class AttachableGeometryField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.AttachableGeometryField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT Attachable_fk, identifier, jsonPath, shortName
            FROM AttachableGeometryField
            WHERE AttachableGeometryField_pk = ?
            ''',
            (self.AttachableGeometryField_pk,)
        ).fetchone()

    @property
    def Attachable_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def shortName(self) -> str:
        return self.query_result()[3]


class AttachableRenderControllerField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.AttachableRenderControllerField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT Attachable_fk, condition, identifier, jsonPath
            FROM AttachableRenderControllerField
            WHERE AttachableRenderControllerField_pk = ?
            ''',
            (self.AttachableRenderControllerField_pk,)
        ).fetchone()

    @property
    def Attachable_fk(self) -> int:
        return self.query_result()[0]

    @property
    def condition(self) -> str | None:
        return self.query_result()[1]

    @property
    def identifier(self) -> str:
        return self.query_result()[2]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[3]


class AttachableAnimationField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.AttachableAnimationField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT Attachable_fk, identifier, jsonPath, shortName
            FROM AttachableAnimationField
            WHERE AttachableAnimationField_pk = ?
            ''',
            (self.AttachableAnimationField_pk,)
        ).fetchone()

    @property
    def Attachable_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def shortName(self) -> str:
        return self.query_result()[3]


class AttachableAnimationControllerField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.AttachableAnimationControllerField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT Attachable_fk, identifier, jsonPath, shortName
            FROM AttachableAnimationControllerField
            WHERE AttachableAnimationControllerField_pk = ?
            ''',
            (self.AttachableAnimationControllerField_pk,)
        ).fetchone()

    @property
    def Attachable_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]

    @property
    def shortName(self) -> str:
        return self.query_result()[3]


class SoundDefinitionsFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.SoundDefinitionsFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ResourcePack_fk, path
            FROM SoundDefinitionsFile
            WHERE SoundDefinitionsFile_pk = ?
            ''',
            (self.SoundDefinitionsFile_pk,)
        ).fetchone()

    @property
    def ResourcePack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class SoundDefinition:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.SoundDefinition_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT SoundDefinitionsFile_fk, identifier, jsonPath
            FROM SoundDefinition
            WHERE SoundDefinition_pk = ?
            ''',
            (self.SoundDefinition_pk,)
        ).fetchone()

    @property
    def SoundDefinitionsFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]


class SoundDefinitionSoundField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.SoundDefinitionSoundField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT SoundDefinition_fk, identifier, jsonPath
            FROM SoundDefinitionSoundField
            WHERE SoundDefinitionSoundField_pk = ?
            ''',
            (self.SoundDefinitionSoundField_pk,)
        ).fetchone()

    @property
    def SoundDefinition_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]


class SoundFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.SoundFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ResourcePack_fk, identifier, path
            FROM SoundFile
            WHERE SoundFile_pk = ?
            ''',
            (self.SoundFile_pk,)
        ).fetchone()

    @property
    def ResourcePack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def path(self) -> Path:
        return self.query_result()[2]


class RpItemFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RpItemFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT ResourcePack_fk, path
            FROM RpItemFile
            WHERE RpItemFile_pk = ?
            ''',
            (self.RpItemFile_pk,)
        ).fetchone()

    @property
    def ResourcePack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class RpItem:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.RpItem_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT RpItemFile_fk, icon, identifier
            FROM RpItem
            WHERE RpItem_pk = ?
            ''',
            (self.RpItem_pk,)
        ).fetchone()

    @property
    def RpItemFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def icon(self) -> str | None:
        return self.query_result()[1]

    @property
    def identifier(self) -> str:
        return self.query_result()[2]


class BehaviorPack:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.BehaviorPack_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT path
            FROM BehaviorPack
            WHERE BehaviorPack_pk = ?
            ''',
            (self.BehaviorPack_pk,)
        ).fetchone()

    @property
    def path(self) -> Path:
        return self.query_result()[0]


class EntityFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.EntityFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT BehaviorPack_fk, path
            FROM EntityFile
            WHERE EntityFile_pk = ?
            ''',
            (self.EntityFile_pk,)
        ).fetchone()

    @property
    def BehaviorPack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class Entity:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.Entity_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT EntityFile_fk, identifier
            FROM Entity
            WHERE Entity_pk = ?
            ''',
            (self.Entity_pk,)
        ).fetchone()

    @property
    def EntityFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]


class EntityLootFieldComponentTypeEnum:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.value: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT 
            FROM EntityLootFieldComponentTypeEnum
            WHERE value = ?
            ''',
            (self.value,)
        ).fetchone()


class EntityLootField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.EntityLootField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT Entity_fk, componentType, identifier, jsonPath
            FROM EntityLootField
            WHERE EntityLootField_pk = ?
            ''',
            (self.EntityLootField_pk,)
        ).fetchone()

    @property
    def Entity_fk(self) -> int:
        return self.query_result()[0]

    @property
    def componentType(self) -> str:
        return self.query_result()[1]

    @property
    def identifier(self) -> str:
        return self.query_result()[2]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[3]


class EntityTradeFieldComponentTypeEnum:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.value: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT 
            FROM EntityTradeFieldComponentTypeEnum
            WHERE value = ?
            ''',
            (self.value,)
        ).fetchone()


class EntityTradeField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.EntityTradeField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT Entity_fk, componentType, identifier, jsonPath
            FROM EntityTradeField
            WHERE EntityTradeField_pk = ?
            ''',
            (self.EntityTradeField_pk,)
        ).fetchone()

    @property
    def Entity_fk(self) -> int:
        return self.query_result()[0]

    @property
    def componentType(self) -> str:
        return self.query_result()[1]

    @property
    def identifier(self) -> str:
        return self.query_result()[2]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[3]


class LootTableFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.LootTableFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT BehaviorPack_fk, path
            FROM LootTableFile
            WHERE LootTableFile_pk = ?
            ''',
            (self.LootTableFile_pk,)
        ).fetchone()

    @property
    def BehaviorPack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class LootTable:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.LootTable_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT LootTableFile_fk, identifier
            FROM LootTable
            WHERE LootTable_pk = ?
            ''',
            (self.LootTable_pk,)
        ).fetchone()

    @property
    def LootTableFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]


class LootTableItemField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.LootTableItemField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT LootTable_fk, identifier, jsonPath
            FROM LootTableItemField
            WHERE LootTableItemField_pk = ?
            ''',
            (self.LootTableItemField_pk,)
        ).fetchone()

    @property
    def LootTable_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]


class LootTableLootTableField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.LootTableLootTableField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT LootTable_fk, identifier, jsonPath
            FROM LootTableLootTableField
            WHERE LootTableLootTableField_pk = ?
            ''',
            (self.LootTableLootTableField_pk,)
        ).fetchone()

    @property
    def LootTable_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]


class TradeTableFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.TradeTableFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT BehaviorPack_fk, path
            FROM TradeTableFile
            WHERE TradeTableFile_pk = ?
            ''',
            (self.TradeTableFile_pk,)
        ).fetchone()

    @property
    def BehaviorPack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class TradeTable:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.TradeTable_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT TradeTableFile_fk, identifier
            FROM TradeTable
            WHERE TradeTable_pk = ?
            ''',
            (self.TradeTable_pk,)
        ).fetchone()

    @property
    def TradeTableFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]


class TradeTableItemField:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.TradeTableItemField_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT TradeTable_fk, dataValue, identifier, jsonPath
            FROM TradeTableItemField
            WHERE TradeTableItemField_pk = ?
            ''',
            (self.TradeTableItemField_pk,)
        ).fetchone()

    @property
    def TradeTable_fk(self) -> int:
        return self.query_result()[0]

    @property
    def dataValue(self) -> int | None:
        return self.query_result()[1]

    @property
    def identifier(self) -> str:
        return self.query_result()[2]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[3]


class BpAnimationFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.BpAnimationFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT BehaviorPack_fk, path
            FROM BpAnimationFile
            WHERE BpAnimationFile_pk = ?
            ''',
            (self.BpAnimationFile_pk,)
        ).fetchone()

    @property
    def BehaviorPack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class BpAnimation:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.BpAnimation_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT BpAnimationFile_fk, identifier, jsonPath
            FROM BpAnimation
            WHERE BpAnimation_pk = ?
            ''',
            (self.BpAnimation_pk,)
        ).fetchone()

    @property
    def BpAnimationFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]


class BpAnimationControllerFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.BpAnimationControllerFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT BehaviorPack_fk, path
            FROM BpAnimationControllerFile
            WHERE BpAnimationControllerFile_pk = ?
            ''',
            (self.BpAnimationControllerFile_pk,)
        ).fetchone()

    @property
    def BehaviorPack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class BpAnimationController:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.BpAnimationController_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT BpAnimationControllerFile_fk, identifier, jsonPath
            FROM BpAnimationController
            WHERE BpAnimationController_pk = ?
            ''',
            (self.BpAnimationController_pk,)
        ).fetchone()

    @property
    def BpAnimationControllerFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def jsonPath(self) -> str:
        return self.query_result()[2]


class BpItemFile:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.BpItemFile_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT BehaviorPack_fk, path
            FROM BpItemFile
            WHERE BpItemFile_pk = ?
            ''',
            (self.BpItemFile_pk,)
        ).fetchone()

    @property
    def BehaviorPack_fk(self) -> int | None:
        return self.query_result()[0]

    @property
    def path(self) -> Path:
        return self.query_result()[1]


class BpItemParserVersionEnum:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.value: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT 
            FROM BpItemParserVersionEnum
            WHERE value = ?
            ''',
            (self.value,)
        ).fetchone()


class BpItem:
    def __init__(self, db: Connection, id: int):
        self.db: Connection = db
        self.BpItem_pk: int = id

    @cache
    def query_result(self):
        return self.db.execute(
            '''
            SELECT BpItemFile_fk, identifier, parserVersion, texture
            FROM BpItem
            WHERE BpItem_pk = ?
            ''',
            (self.BpItem_pk,)
        ).fetchone()

    @property
    def BpItemFile_fk(self) -> int:
        return self.query_result()[0]

    @property
    def identifier(self) -> str:
        return self.query_result()[1]

    @property
    def parserVersion(self) -> str:
        return self.query_result()[2]

    @property
    def texture(self) -> str | None:
        return self.query_result()[3]


# Map strings to class names for easy access
WRAPPER_CLASSES = {
    'ResourcePack': ResourcePack,
    'ClientEntityFile': ClientEntityFile,
    'ClientEntity': ClientEntity,
    'ClientEntityRenderControllerField': ClientEntityRenderControllerField,
    'ClientEntityGeometryField': ClientEntityGeometryField,
    'ClientEntityTextureField': ClientEntityTextureField,
    'ClientEntityMaterialField': ClientEntityMaterialField,
    'ClientEntityAnimationField': ClientEntityAnimationField,
    'ClientEntityAnimationControllerField': ClientEntityAnimationControllerField,
    'RenderControllerFile': RenderControllerFile,
    'RenderController': RenderController,
    'RenderControllerTexturesField': RenderControllerTexturesField,
    'RenderControllerMaterialsField': RenderControllerMaterialsField,
    'RenderControllerGeometryField': RenderControllerGeometryField,
    'GeometryFile': GeometryFile,
    'Geometry': Geometry,
    'TextureFile': TextureFile,
    'ParticleFile': ParticleFile,
    'Particle': Particle,
    'RpAnimationFile': RpAnimationFile,
    'RpAnimation': RpAnimation,
    'RpAnimationParticleEffect': RpAnimationParticleEffect,
    'RpAnimationSoundEffect': RpAnimationSoundEffect,
    'RpAnimationControllerFile': RpAnimationControllerFile,
    'RpAnimationController': RpAnimationController,
    'RpAnimationControllerParticleEffect': RpAnimationControllerParticleEffect,
    'RpAnimationControllerSoundEffect': RpAnimationControllerSoundEffect,
    'AttachableFile': AttachableFile,
    'Attachable': Attachable,
    'AttachableItemField': AttachableItemField,
    'AttachableMaterialField': AttachableMaterialField,
    'AttachableTextureField': AttachableTextureField,
    'AttachableGeometryField': AttachableGeometryField,
    'AttachableRenderControllerField': AttachableRenderControllerField,
    'AttachableAnimationField': AttachableAnimationField,
    'AttachableAnimationControllerField': AttachableAnimationControllerField,
    'SoundDefinitionsFile': SoundDefinitionsFile,
    'SoundDefinition': SoundDefinition,
    'SoundDefinitionSoundField': SoundDefinitionSoundField,
    'SoundFile': SoundFile,
    'RpItemFile': RpItemFile,
    'RpItem': RpItem,
    'BehaviorPack': BehaviorPack,
    'EntityFile': EntityFile,
    'Entity': Entity,
    'EntityLootFieldComponentTypeEnum': EntityLootFieldComponentTypeEnum,
    'EntityLootField': EntityLootField,
    'EntityTradeFieldComponentTypeEnum': EntityTradeFieldComponentTypeEnum,
    'EntityTradeField': EntityTradeField,
    'LootTableFile': LootTableFile,
    'LootTable': LootTable,
    'LootTableItemField': LootTableItemField,
    'LootTableLootTableField': LootTableLootTableField,
    'TradeTableFile': TradeTableFile,
    'TradeTable': TradeTable,
    'TradeTableItemField': TradeTableItemField,
    'BpAnimationFile': BpAnimationFile,
    'BpAnimation': BpAnimation,
    'BpAnimationControllerFile': BpAnimationControllerFile,
    'BpAnimationController': BpAnimationController,
    'BpItemFile': BpItemFile,
    'BpItemParserVersionEnum': BpItemParserVersionEnum,
    'BpItem': BpItem,
}


# Weird "import X as X" syntax is for some IDEs to recognize the imports
# as values accessible from the module.
from .db_main import (
    Database as Database
)
from .easy_query import (
    EasyQuery as EasyQuery,
    Left as Left,
)

VERSION = (2, 0, 0)  # COMPATIBILITY BREAK, NEW FEATURE, BUGFIX
__version__ = '.'.join([str(x) for x in VERSION])



"""
THIS FILE IS GENERATED. DO NOT EDIT IT MANUALLY.
"""
from __future__ import annotations
from typing import NamedTuple

class _TableConnection(NamedTuple):
    columns: tuple[str, str]
    is_pk: bool

RELATION_MAP = {
    "ClientEntity": {
        "Entity": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "ClientEntityFile": _TableConnection(
            columns=(('ClientEntityFile_fk', 'ClientEntityFile_pk')),
            is_pk=True
        ),
        "ClientEntityAnimationControllerField": _TableConnection(
            columns=(('ClientEntity_pk', 'ClientEntity_fk')),
            is_pk=True
        ),
        "ClientEntityAnimationField": _TableConnection(
            columns=(('ClientEntity_pk', 'ClientEntity_fk')),
            is_pk=True
        ),
        "ClientEntityGeometryField": _TableConnection(
            columns=(('ClientEntity_pk', 'ClientEntity_fk')),
            is_pk=True
        ),
        "ClientEntityMaterialField": _TableConnection(
            columns=(('ClientEntity_pk', 'ClientEntity_fk')),
            is_pk=True
        ),
        "ClientEntityRenderControllerField": _TableConnection(
            columns=(('ClientEntity_pk', 'ClientEntity_fk')),
            is_pk=True
        ),
        "ClientEntityTextureField": _TableConnection(
            columns=(('ClientEntity_pk', 'ClientEntity_fk')),
            is_pk=True
        ),
    },
    "ClientEntityRenderControllerField": {
        "RenderController": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "ClientEntity": _TableConnection(
            columns=(('ClientEntity_fk', 'ClientEntity_pk')),
            is_pk=True
        ),
    },
    "ClientEntityGeometryField": {
        "Geometry": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "ClientEntity": _TableConnection(
            columns=(('ClientEntity_fk', 'ClientEntity_pk')),
            is_pk=True
        ),
    },
    "ClientEntityTextureField": {
        "TextureFile": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "ClientEntity": _TableConnection(
            columns=(('ClientEntity_fk', 'ClientEntity_pk')),
            is_pk=True
        ),
    },
    "ClientEntityAnimationField": {
        "RpAnimation": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "ClientEntity": _TableConnection(
            columns=(('ClientEntity_fk', 'ClientEntity_pk')),
            is_pk=True
        ),
    },
    "ClientEntityAnimationControllerField": {
        "RpAnimationController": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "ClientEntity": _TableConnection(
            columns=(('ClientEntity_fk', 'ClientEntity_pk')),
            is_pk=True
        ),
    },
    "AttachableItemField": {
        "RpItem": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "BpItem": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "Attachable": _TableConnection(
            columns=(('Attachable_fk', 'Attachable_pk')),
            is_pk=True
        ),
    },
    "AttachableTextureField": {
        "TextureFile": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "Attachable": _TableConnection(
            columns=(('Attachable_fk', 'Attachable_pk')),
            is_pk=True
        ),
    },
    "AttachableGeometryField": {
        "Geometry": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "Attachable": _TableConnection(
            columns=(('Attachable_fk', 'Attachable_pk')),
            is_pk=True
        ),
    },
    "AttachableRenderControllerField": {
        "RenderController": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "Attachable": _TableConnection(
            columns=(('Attachable_fk', 'Attachable_pk')),
            is_pk=True
        ),
    },
    "AttachableAnimationField": {
        "RpAnimation": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "Attachable": _TableConnection(
            columns=(('Attachable_fk', 'Attachable_pk')),
            is_pk=True
        ),
    },
    "AttachableAnimationControllerField": {
        "RpAnimationController": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "Attachable": _TableConnection(
            columns=(('Attachable_fk', 'Attachable_pk')),
            is_pk=True
        ),
    },
    "EntityLootField": {
        "LootTable": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "EntityLootFieldComponentTypeEnum": _TableConnection(
            columns=(('componentType', 'value')),
            is_pk=True
        ),
        "Entity": _TableConnection(
            columns=(('Entity_fk', 'Entity_pk')),
            is_pk=True
        ),
    },
    "EntityTradeField": {
        "TradeTable": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "EntityTradeFieldComponentTypeEnum": _TableConnection(
            columns=(('componentType', 'value')),
            is_pk=True
        ),
        "Entity": _TableConnection(
            columns=(('Entity_fk', 'Entity_pk')),
            is_pk=True
        ),
    },
    "LootTableItemField": {
        "RpItem": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "BpItem": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "LootTable": _TableConnection(
            columns=(('LootTable_fk', 'LootTable_pk')),
            is_pk=True
        ),
    },
    "LootTableLootTableField": {
        "LootTable": _TableConnection(
            columns=(('LootTable_fk', 'LootTable_pk')),
            is_pk=True
        ),
    },
    "TradeTableItemField": {
        "RpItem": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "BpItem": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "TradeTable": _TableConnection(
            columns=(('TradeTable_fk', 'TradeTable_pk')),
            is_pk=True
        ),
    },
    "Particle": {
        "TextureFile": _TableConnection(
            columns=(('texture', 'identifier')),
            is_pk=False
        ),
        "ParticleFile": _TableConnection(
            columns=(('ParticleFile_fk', 'ParticleFile_pk')),
            is_pk=True
        ),
    },
    "SoundDefinitionSoundField": {
        "SoundFile": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "SoundDefinition": _TableConnection(
            columns=(('SoundDefinition_fk', 'SoundDefinition_pk')),
            is_pk=True
        ),
    },
    "Entity": {
        "ClientEntity": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "EntityFile": _TableConnection(
            columns=(('EntityFile_fk', 'EntityFile_pk')),
            is_pk=True
        ),
        "EntityLootField": _TableConnection(
            columns=(('Entity_pk', 'Entity_fk')),
            is_pk=True
        ),
        "EntityTradeField": _TableConnection(
            columns=(('Entity_pk', 'Entity_fk')),
            is_pk=True
        ),
    },
    "RenderController": {
        "ClientEntityRenderControllerField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "AttachableRenderControllerField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "RenderControllerFile": _TableConnection(
            columns=(('RenderControllerFile_fk', 'RenderControllerFile_pk')),
            is_pk=True
        ),
        "RenderControllerGeometryField": _TableConnection(
            columns=(('RenderController_pk', 'RenderController_fk')),
            is_pk=True
        ),
        "RenderControllerMaterialsField": _TableConnection(
            columns=(('RenderController_pk', 'RenderController_fk')),
            is_pk=True
        ),
        "RenderControllerTexturesField": _TableConnection(
            columns=(('RenderController_pk', 'RenderController_fk')),
            is_pk=True
        ),
    },
    "Geometry": {
        "ClientEntityGeometryField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "AttachableGeometryField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "GeometryFile": _TableConnection(
            columns=(('GeometryFile_fk', 'GeometryFile_pk')),
            is_pk=True
        ),
    },
    "TextureFile": {
        "ClientEntityTextureField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "AttachableTextureField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "Particle": _TableConnection(
            columns=(('identifier', 'texture')),
            is_pk=False
        ),
        "ResourcePack": _TableConnection(
            columns=(('ResourcePack_fk', 'ResourcePack_pk')),
            is_pk=True
        ),
    },
    "RpAnimation": {
        "ClientEntityAnimationField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "AttachableAnimationField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "RpAnimationFile": _TableConnection(
            columns=(('RpAnimationFile_fk', 'RpAnimationFile_pk')),
            is_pk=True
        ),
        "RpAnimationParticleEffect": _TableConnection(
            columns=(('RpAnimation_pk', 'RpAnimation_fk')),
            is_pk=True
        ),
        "RpAnimationSoundEffect": _TableConnection(
            columns=(('RpAnimation_pk', 'RpAnimation_fk')),
            is_pk=True
        ),
    },
    "RpAnimationController": {
        "ClientEntityAnimationControllerField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "AttachableAnimationControllerField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "RpAnimationControllerFile": _TableConnection(
            columns=(('RpAnimationControllerFile_fk', 'RpAnimationControllerFile_pk')),
            is_pk=True
        ),
        "RpAnimationControllerParticleEffect": _TableConnection(
            columns=(('RpAnimationController_pk', 'RpAnimationController_fk')),
            is_pk=True
        ),
        "RpAnimationControllerSoundEffect": _TableConnection(
            columns=(('RpAnimationController_pk', 'RpAnimationController_fk')),
            is_pk=True
        ),
    },
    "RpItem": {
        "AttachableItemField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "LootTableItemField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "TradeTableItemField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "RpItemFile": _TableConnection(
            columns=(('RpItemFile_fk', 'RpItemFile_pk')),
            is_pk=True
        ),
    },
    "BpItem": {
        "AttachableItemField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "LootTableItemField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "TradeTableItemField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "BpItemParserVersionEnum": _TableConnection(
            columns=(('parserVersion', 'value')),
            is_pk=True
        ),
        "BpItemFile": _TableConnection(
            columns=(('BpItemFile_fk', 'BpItemFile_pk')),
            is_pk=True
        ),
    },
    "LootTable": {
        "EntityLootField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "LootTableLootTableField": _TableConnection(
            columns=(('LootTable_pk', 'LootTable_fk')),
            is_pk=True
        ),
        "LootTableFile": _TableConnection(
            columns=(('LootTableFile_fk', 'LootTableFile_pk')),
            is_pk=True
        ),
        "LootTableItemField": _TableConnection(
            columns=(('LootTable_pk', 'LootTable_fk')),
            is_pk=True
        ),
    },
    "TradeTable": {
        "EntityTradeField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "TradeTableFile": _TableConnection(
            columns=(('TradeTableFile_fk', 'TradeTableFile_pk')),
            is_pk=True
        ),
        "TradeTableItemField": _TableConnection(
            columns=(('TradeTable_pk', 'TradeTable_fk')),
            is_pk=True
        ),
    },
    "SoundFile": {
        "SoundDefinitionSoundField": _TableConnection(
            columns=(('identifier', 'identifier')),
            is_pk=False
        ),
        "ResourcePack": _TableConnection(
            columns=(('ResourcePack_fk', 'ResourcePack_pk')),
            is_pk=True
        ),
    },
    "AttachableFile": {
        "Attachable": _TableConnection(
            columns=(('AttachableFile_pk', 'AttachableFile_fk')),
            is_pk=True
        ),
        "ResourcePack": _TableConnection(
            columns=(('ResourcePack_fk', 'ResourcePack_pk')),
            is_pk=True
        ),
    },
    "Attachable": {
        "AttachableFile": _TableConnection(
            columns=(('AttachableFile_fk', 'AttachableFile_pk')),
            is_pk=True
        ),
        "AttachableAnimationControllerField": _TableConnection(
            columns=(('Attachable_pk', 'Attachable_fk')),
            is_pk=True
        ),
        "AttachableAnimationField": _TableConnection(
            columns=(('Attachable_pk', 'Attachable_fk')),
            is_pk=True
        ),
        "AttachableGeometryField": _TableConnection(
            columns=(('Attachable_pk', 'Attachable_fk')),
            is_pk=True
        ),
        "AttachableItemField": _TableConnection(
            columns=(('Attachable_pk', 'Attachable_fk')),
            is_pk=True
        ),
        "AttachableMaterialField": _TableConnection(
            columns=(('Attachable_pk', 'Attachable_fk')),
            is_pk=True
        ),
        "AttachableRenderControllerField": _TableConnection(
            columns=(('Attachable_pk', 'Attachable_fk')),
            is_pk=True
        ),
        "AttachableTextureField": _TableConnection(
            columns=(('Attachable_pk', 'Attachable_fk')),
            is_pk=True
        ),
    },
    "ResourcePack": {
        "AttachableFile": _TableConnection(
            columns=(('ResourcePack_pk', 'ResourcePack_fk')),
            is_pk=True
        ),
        "ClientEntityFile": _TableConnection(
            columns=(('ResourcePack_pk', 'ResourcePack_fk')),
            is_pk=True
        ),
        "GeometryFile": _TableConnection(
            columns=(('ResourcePack_pk', 'ResourcePack_fk')),
            is_pk=True
        ),
        "ParticleFile": _TableConnection(
            columns=(('ResourcePack_pk', 'ResourcePack_fk')),
            is_pk=True
        ),
        "RenderControllerFile": _TableConnection(
            columns=(('ResourcePack_pk', 'ResourcePack_fk')),
            is_pk=True
        ),
        "RpAnimationControllerFile": _TableConnection(
            columns=(('ResourcePack_pk', 'ResourcePack_fk')),
            is_pk=True
        ),
        "RpAnimationFile": _TableConnection(
            columns=(('ResourcePack_pk', 'ResourcePack_fk')),
            is_pk=True
        ),
        "RpItemFile": _TableConnection(
            columns=(('ResourcePack_pk', 'ResourcePack_fk')),
            is_pk=True
        ),
        "SoundDefinitionsFile": _TableConnection(
            columns=(('ResourcePack_pk', 'ResourcePack_fk')),
            is_pk=True
        ),
        "SoundFile": _TableConnection(
            columns=(('ResourcePack_pk', 'ResourcePack_fk')),
            is_pk=True
        ),
        "TextureFile": _TableConnection(
            columns=(('ResourcePack_pk', 'ResourcePack_fk')),
            is_pk=True
        ),
    },
    "AttachableMaterialField": {
        "Attachable": _TableConnection(
            columns=(('Attachable_fk', 'Attachable_pk')),
            is_pk=True
        ),
    },
    "BpAnimationFile": {
        "BpAnimation": _TableConnection(
            columns=(('BpAnimationFile_pk', 'BpAnimationFile_fk')),
            is_pk=True
        ),
        "BehaviorPack": _TableConnection(
            columns=(('BehaviorPack_fk', 'BehaviorPack_pk')),
            is_pk=True
        ),
    },
    "BpAnimation": {
        "BpAnimationFile": _TableConnection(
            columns=(('BpAnimationFile_fk', 'BpAnimationFile_pk')),
            is_pk=True
        ),
    },
    "BpAnimationControllerFile": {
        "BpAnimationController": _TableConnection(
            columns=(('BpAnimationControllerFile_pk', 'BpAnimationControllerFile_fk')),
            is_pk=True
        ),
        "BehaviorPack": _TableConnection(
            columns=(('BehaviorPack_fk', 'BehaviorPack_pk')),
            is_pk=True
        ),
    },
    "BpAnimationController": {
        "BpAnimationControllerFile": _TableConnection(
            columns=(('BpAnimationControllerFile_fk', 'BpAnimationControllerFile_pk')),
            is_pk=True
        ),
    },
    "BehaviorPack": {
        "BpAnimationControllerFile": _TableConnection(
            columns=(('BehaviorPack_pk', 'BehaviorPack_fk')),
            is_pk=True
        ),
        "BpAnimationFile": _TableConnection(
            columns=(('BehaviorPack_pk', 'BehaviorPack_fk')),
            is_pk=True
        ),
        "BpItemFile": _TableConnection(
            columns=(('BehaviorPack_pk', 'BehaviorPack_fk')),
            is_pk=True
        ),
        "EntityFile": _TableConnection(
            columns=(('BehaviorPack_pk', 'BehaviorPack_fk')),
            is_pk=True
        ),
        "LootTableFile": _TableConnection(
            columns=(('BehaviorPack_pk', 'BehaviorPack_fk')),
            is_pk=True
        ),
        "TradeTableFile": _TableConnection(
            columns=(('BehaviorPack_pk', 'BehaviorPack_fk')),
            is_pk=True
        ),
    },
    "BpItemParserVersionEnum": {
        "BpItem": _TableConnection(
            columns=(('value', 'parserVersion')),
            is_pk=True
        ),
    },
    "BpItemFile": {
        "BpItem": _TableConnection(
            columns=(('BpItemFile_pk', 'BpItemFile_fk')),
            is_pk=True
        ),
        "BehaviorPack": _TableConnection(
            columns=(('BehaviorPack_fk', 'BehaviorPack_pk')),
            is_pk=True
        ),
    },
    "ClientEntityFile": {
        "ClientEntity": _TableConnection(
            columns=(('ClientEntityFile_pk', 'ClientEntityFile_fk')),
            is_pk=True
        ),
        "ResourcePack": _TableConnection(
            columns=(('ResourcePack_fk', 'ResourcePack_pk')),
            is_pk=True
        ),
    },
    "ClientEntityMaterialField": {
        "ClientEntity": _TableConnection(
            columns=(('ClientEntity_fk', 'ClientEntity_pk')),
            is_pk=True
        ),
    },
    "EntityFile": {
        "Entity": _TableConnection(
            columns=(('EntityFile_pk', 'EntityFile_fk')),
            is_pk=True
        ),
        "BehaviorPack": _TableConnection(
            columns=(('BehaviorPack_fk', 'BehaviorPack_pk')),
            is_pk=True
        ),
    },
    "EntityLootFieldComponentTypeEnum": {
        "EntityLootField": _TableConnection(
            columns=(('value', 'componentType')),
            is_pk=True
        ),
    },
    "EntityTradeFieldComponentTypeEnum": {
        "EntityTradeField": _TableConnection(
            columns=(('value', 'componentType')),
            is_pk=True
        ),
    },
    "GeometryFile": {
        "Geometry": _TableConnection(
            columns=(('GeometryFile_pk', 'GeometryFile_fk')),
            is_pk=True
        ),
        "ResourcePack": _TableConnection(
            columns=(('ResourcePack_fk', 'ResourcePack_pk')),
            is_pk=True
        ),
    },
    "LootTableFile": {
        "LootTable": _TableConnection(
            columns=(('LootTableFile_pk', 'LootTableFile_fk')),
            is_pk=True
        ),
        "BehaviorPack": _TableConnection(
            columns=(('BehaviorPack_fk', 'BehaviorPack_pk')),
            is_pk=True
        ),
    },
    "ParticleFile": {
        "Particle": _TableConnection(
            columns=(('ParticleFile_pk', 'ParticleFile_fk')),
            is_pk=True
        ),
        "ResourcePack": _TableConnection(
            columns=(('ResourcePack_fk', 'ResourcePack_pk')),
            is_pk=True
        ),
    },
    "RenderControllerFile": {
        "RenderController": _TableConnection(
            columns=(('RenderControllerFile_pk', 'RenderControllerFile_fk')),
            is_pk=True
        ),
        "ResourcePack": _TableConnection(
            columns=(('ResourcePack_fk', 'ResourcePack_pk')),
            is_pk=True
        ),
    },
    "RenderControllerGeometryField": {
        "RenderController": _TableConnection(
            columns=(('RenderController_fk', 'RenderController_pk')),
            is_pk=True
        ),
    },
    "RenderControllerMaterialsField": {
        "RenderController": _TableConnection(
            columns=(('RenderController_fk', 'RenderController_pk')),
            is_pk=True
        ),
    },
    "RenderControllerTexturesField": {
        "RenderController": _TableConnection(
            columns=(('RenderController_fk', 'RenderController_pk')),
            is_pk=True
        ),
    },
    "RpAnimationFile": {
        "RpAnimation": _TableConnection(
            columns=(('RpAnimationFile_pk', 'RpAnimationFile_fk')),
            is_pk=True
        ),
        "ResourcePack": _TableConnection(
            columns=(('ResourcePack_fk', 'ResourcePack_pk')),
            is_pk=True
        ),
    },
    "RpAnimationControllerFile": {
        "RpAnimationController": _TableConnection(
            columns=(('RpAnimationControllerFile_pk', 'RpAnimationControllerFile_fk')),
            is_pk=True
        ),
        "ResourcePack": _TableConnection(
            columns=(('ResourcePack_fk', 'ResourcePack_pk')),
            is_pk=True
        ),
    },
    "RpAnimationControllerParticleEffect": {
        "RpAnimationController": _TableConnection(
            columns=(('RpAnimationController_fk', 'RpAnimationController_pk')),
            is_pk=True
        ),
    },
    "RpAnimationControllerSoundEffect": {
        "RpAnimationController": _TableConnection(
            columns=(('RpAnimationController_fk', 'RpAnimationController_pk')),
            is_pk=True
        ),
    },
    "RpAnimationParticleEffect": {
        "RpAnimation": _TableConnection(
            columns=(('RpAnimation_fk', 'RpAnimation_pk')),
            is_pk=True
        ),
    },
    "RpAnimationSoundEffect": {
        "RpAnimation": _TableConnection(
            columns=(('RpAnimation_fk', 'RpAnimation_pk')),
            is_pk=True
        ),
    },
    "RpItemFile": {
        "RpItem": _TableConnection(
            columns=(('RpItemFile_pk', 'RpItemFile_fk')),
            is_pk=True
        ),
        "ResourcePack": _TableConnection(
            columns=(('ResourcePack_fk', 'ResourcePack_pk')),
            is_pk=True
        ),
    },
    "SoundDefinitionsFile": {
        "SoundDefinition": _TableConnection(
            columns=(('SoundDefinitionsFile_pk', 'SoundDefinitionsFile_fk')),
            is_pk=True
        ),
        "ResourcePack": _TableConnection(
            columns=(('ResourcePack_fk', 'ResourcePack_pk')),
            is_pk=True
        ),
    },
    "SoundDefinition": {
        "SoundDefinitionsFile": _TableConnection(
            columns=(('SoundDefinitionsFile_fk', 'SoundDefinitionsFile_pk')),
            is_pk=True
        ),
        "SoundDefinitionSoundField": _TableConnection(
            columns=(('SoundDefinition_pk', 'SoundDefinition_fk')),
            is_pk=True
        ),
    },
    "TradeTableFile": {
        "TradeTable": _TableConnection(
            columns=(('TradeTableFile_pk', 'TradeTableFile_fk')),
            is_pk=True
        ),
        "BehaviorPack": _TableConnection(
            columns=(('BehaviorPack_fk', 'BehaviorPack_pk')),
            is_pk=True
        ),
    },
}


from __future__ import annotations
from sqlite3 import Connection, Cursor
from typing import Iterable, Iterator, NamedTuple, Any, TYPE_CHECKING
from dataclasses import dataclass
from collections import deque
if TYPE_CHECKING:
    from .db_main import Database


class Left(NamedTuple):
    '''
    A helper class to indicate that a table should be joined using LEFT join in
    the query of :class:`EasyQuery` class.
    '''
    value: str
    '''The name of the table'''

@dataclass
class EasyQuery:
    '''
    EasyQuery is a class which allows quick and easy way of building and
    executing queries on the database.

    In most cases queries should be build using the :meth:`EasyQuery.build`
    method. Creating them manually is possible and can be useful if you want to
    use benefits of some of the methods of this class. In this case make sure
    that the results of the query are pimary keys of the tables and that the
    columns are named after the tables.
    '''
    connection: Connection
    '''The database connection.'''

    sql_code: str
    '''The query runnned by this instance of :class:`EasyQuery`.'''

    @staticmethod
    def build(
            db: Connection | Database | None, root: str, *tables: str | Left,
            blacklist: Iterable[str] = ("BehaviorPack", "ResourcePack"),
            accept_non_pk: bool = True, distinct: bool = True,
            where: list[str] | None = None,
            group_by: list[str] | None = None,
            having: list[str] | None = None,
            order_by: list[str] | None = None) -> EasyQuery:
        '''
        Creates an instance of :class:`EasyQuery` from the given properties.
        This is a go-to method for creating an instance of :class:`EasyQuery`.

        This function automatically finds the relations in the database to find
        the connections between provided tables. The rows of the results of the
        query contain primary keys of the provided tables. The query build by
        this function is then passed to the :class:`EasyQuery` constructor.

        Example:

        .. code-block:: python

            >>> query = EasyQuery(
            ...     None, "Entity", Left("Geometry"), "RpAnimation",
            ...     accept_non_pk=True,
            ...     where=[
            ...         "EntityFile.EntityFile_pk == 1"
            ...     ]
            ... ).sql_code
            >>> print(query)
            SELECT DISTINCT
                    Entity_pk AS Entity,
                    Geometry_pk AS Geometry,
                    RpAnimation_pk AS RpAnimation
            FROM Entity
            JOIN ClientEntity
                    ON Entity.identifier = ClientEntity.identifier
            JOIN ClientEntityGeometryField
                    ON ClientEntity.ClientEntity_pk = ClientEntityGeometryField.ClientEntity_fk
            LEFT JOIN Geometry
                    ON ClientEntityGeometryField.identifier = Geometry.identifier
            JOIN ClientEntityAnimationField
                    ON ClientEntity.ClientEntity_pk = ClientEntityAnimationField.ClientEntity_fk
            JOIN RpAnimation
                    ON ClientEntityAnimationField.identifier = RpAnimation.identifier
            WHERE
                    EntityFile.EntityFile_pk == 1

        :param db: The database connection or :class:`Database` object. The
            db can be None if you only want to get the query string and don't
            care about running it.
        :param root: The root table to start the query from.
        :param tables: A list of tables to join. Use Left to indicate that a table
            should be joined using LEFT join. The tables don't need to have a
            direct connection between each other, the connections will be found
            automatically if necessary relations exist (this means that the query
            genrated can include tables that are not in the list).
        :param blacklist: A list of tables to ignore while searching for
            connections. By default it's BehaviorPack and ResourcePack because
            otherwise in many cases the query would look for objects from the same
            packs instead of using different more useful connections.
        :param accept_non_pk: Whether to accept non-primary key relations.
            By default it's True. The primary key connections only cover the
            situations where it's guaranteed that the relation is valid (like
            a relation between EntityFile and Entity). Most of the connections
            that might be useful to query are non-primary key connections (like
            a relation between a ClientEntity and RenderController).
        :param distinct: Whether to use DISTINCT in the query. By default it's
            True. It's recommended to use it when querying for multiple tables
            because otherwise the query might return the same row multiple times.
        :param where: A list of constraints to add to the query. This is a
            list of strings with raw SQL code which is inserted into the WHERE
            part of the query. The constraints are joined using AND.
        :param group_by: A list of columns to group the results by. This is a
            list of strings with raw SQL code which is inserted into the GROUP BY
            part of the query.
        :param having: A list of constraints to add to the query. This is a
            list of strings with raw SQL code which is inserted into the HAVING
            part of the query. The constraints are joined using AND.
        :param order_by: A list of columns to order the results by. This is a
            list of strings with raw SQL code which is inserted into the ORDER BY
            part of the query.


        '''
        # If db is None, continue (this is useful for testing)
        if db is not None and not isinstance(db, Connection):
            # isinstance(db, Database) - not checking for Ddatabase because
            # I don't want to run into circular imports in the future. It
            # is however properly type hinted thanks to TYPE_CHECKING.
            db = db.connection
        return EasyQuery(
            db,
            _easy_query(
                root, *tables,
                blacklist=blacklist,
                accept_non_pk=accept_non_pk,
                distinct=distinct,
                where=where,
                group_by=group_by,
                having=having,
                order_by=order_by
            )
        )

    def run(self) -> Cursor:
        '''
        Run the query on the database and return the cursor.
        '''
        return self.connection.execute(self.sql_code)

    def yield_wrappers(self) -> Iterator[tuple]:
        '''
        Returns an iterator that yields the wrapper classes from the query
        results. The results are tuples with wrapper classes from
        :mod:`sqlite_bedrock_packs.wrappers` module based on the tables in
        the query. If the query is using LEFT join, the wrapper class will be
        None if the row doesn't have a value.
        '''
        from .wrappers import WRAPPER_CLASSES
        cursor = self.run()
        wrappers = [
            WRAPPER_CLASSES[d[0]] for d in cursor.description
        ]
        for row in cursor:
            yield tuple(
                None if value is None else wrapper(self.connection, value)
                for value, wrapper in zip(row, wrappers)
            )


@dataclass
class _TableConnection:
    left: str
    left_column: str
    right: str
    right_column: str
    left_join: bool = False


def _easy_query(
        root: str, *tables: str | Left, blacklist: Iterable[str],
        accept_non_pk: bool, distinct: bool, where: list[str] | None,
        group_by: list[str] | None, having: list[str] | None,
        order_by: list[str] | None) -> str:
    '''
    A helper function that builds queries for :class:`EasyQuery` class.
    '''
    from .graph import RELATION_MAP
    all_tables: list[str | Left] = [root, *tables]
    for t in all_tables:
        t_val = t.value if isinstance(t, Left) else t
        if t_val not in RELATION_MAP.keys():
            raise ValueError(
                f"Table '{t_val}' does not exist in the database.")
    prev_t = None
    joined_connections: list[_TableConnection] = []
    blacklist = list(blacklist)
    for t in all_tables:
        if prev_t is None:
            prev_t = t
            continue
        left = False
        if isinstance(t, Left):
            left = True
            t = t.value
        connection = _find_connection(
            prev_t, t, accept_non_pk, set(blacklist))
        if connection is None:
            raise ValueError(
                f"No connection between {prev_t} and {t} after excluding "
                f"tables: {', '.join(blacklist)}")
        # if left and len(connection) > 1:
        if left:
            connection[-1].left_join = True
        joined_connections.extend(connection)
        prev_t = t
    # Strip joined_connections of duplicates
    reduced_joined_connections = []
    if len(joined_connections) > 0:
        known_connections = {joined_connections[0].left}
        for jc in joined_connections:
            if jc.right in known_connections:
                continue
            known_connections.add(jc.right)
            reduced_joined_connections.append(jc)

    # Convert all tables to list[str] (we don't need Left objects anymore)
    # the infromation is in the reduced_joined_queries
    all_tables = [
        t.value if isinstance(t, Left) else t
        for t in all_tables]
    # Build the quer
    selection = ",\n\t".join([f"{t}_pk AS {t}" for t in all_tables])
    select = "SELECT DISTINCT" if distinct else "SELECT"
    query = f'{select}\n\t{selection}\nFROM {all_tables[0]}'
    for c in reduced_joined_connections:
        join = "LEFT JOIN" if c.left_join else "JOIN"
        query += (
            f'\n{join} {c.right}\n'
            f'\tON {c.left}.{c.left_column} = {c.right}.{c.right_column}')
    if where is not None:
        if isinstance(where, str):
            where = [where]
        query += f'\nWHERE\n\t'+"\n\tAND ".join(where)
    if group_by is not None:
        if isinstance(group_by, str):
            group_by = [group_by]
        query += f'\nGROUP BY\n\t'+"\n\t, ".join(group_by)
    if having is not None:
        if isinstance(having, str):
            having = [having]
        query += f'\nHAVING\n\t'+"\n\tAND ".join(having)
    if order_by is not None:
        if isinstance(order_by, str):
            order_by = [order_by]
        query += f'\nORDER BY\n\t'+"\n\t, ".join(order_by)
    return query

def _find_connection(
        start: str, end: str, accept_non_pk: bool, visited: set[str]):
    '''
    Finds a connnection between two tables in the database.

    :param db: The database connection.
    :param start: The starting table.
    :param end: The ending table.
    :param accept_non_pk: Whether to accept non-primary key relations.
    :param visited: A list of tables to ignore while searching.
    '''
    from .graph import RELATION_MAP
    queue: deque[tuple[str, list[_TableConnection]]] = deque([(start, [])])
    while len(queue) > 0:
        # Pop the first element (shortest path visited)
        node, path = queue.popleft()
        if node in visited:
            continue
        visited.add(node)

        # If node is the end, return the path
        if node == end:
            return path

        # Add all connections to the end of the queue
        for child, relation in RELATION_MAP.get(node, {}).items():
            if not accept_non_pk and not relation.is_pk:
                continue
            queue.append((child, path + [
                _TableConnection(
                    left=node,
                    left_column=relation.columns[0],
                    right=child,
                    right_column=relation.columns[1]
                )
            ]))
    return None
